@@ -1,5 +1,5 @@
 /*
- * tio - a simple serial terminal I/O tool
+ * tio - a serial device I/O tool
  *
  * Copyright (c) 2020-2022  Liam Beguin
  * Copyright (c) 2022  Martin Lund
@@ -60,6 +60,9 @@ struct config_t
     char *log_filename;
     char *socket;
     char *map;
+    char *script;
+    char *script_filename;
+    bool script_run;
 };
 
 static struct config_t c;
@@ -210,9 +213,13 @@ static int data_handler(void *user, const char *section, const char *name,
         {
             option.local_echo = read_boolean(value, name);
         }
-        else if (!strcmp(name, "hexadecimal"))
+        else if (!strcmp(name, "input-mode"))
         {
-            option.hex_mode = read_boolean(value, name);
+            option.input_mode = input_mode_option_parse(value);
+        }
+        else if (!strcmp(name, "output-mode"))
+        {
+            option.output_mode = output_mode_option_parse(value);
         }
         else if (!strcmp(name, "timestamp"))
         {
@@ -259,20 +266,16 @@ static int data_handler(void *user, const char *section, const char *name,
         }
         else if (!strcmp(name, "prefix-ctrl-key"))
         {
-            if (ctrl_key_code(value[0]) > 0)
+            if (!strcmp(value, "none"))
+            {
+                option.prefix_enabled = false;
+            }
+            else if (ctrl_key_code(value[0]) > 0)
             {
                 option.prefix_code = ctrl_key_code(value[0]);
                 option.prefix_key = value[0];
             }
         }
-        else if (!strcmp(name, "response-wait"))
-        {
-            option.response_wait = read_boolean(value, name);
-        }
-        else if (!strcmp(name, "response-timeout"))
-        {
-            option.response_timeout = read_integer(value, name, 0, LONG_MAX);
-        }
         else if (!strcmp(name, "rs-485"))
         {
             option.rs485 = read_boolean(value, name);
@@ -293,6 +296,20 @@ static int data_handler(void *user, const char *section, const char *name,
         {
             // Do nothing
         }
+        else if (!strcmp(name, "script"))
+        {
+            asprintf(&c.script, "%s", value);
+            option.script = c.script;
+        }
+        else if (!strcmp(name, "script-file"))
+        {
+            asprintf(&c.script_filename, "%s", value);
+            option.script_filename = c.script_filename;
+        }
+        else if (!strcmp(name, "script-run"))
+        {
+            option.script_run = script_run_option_parse(value);
+        }
         else
         {
             tio_warning_printf("Unknown option '%s' in configuration file, ignored", name);
@@ -376,30 +393,36 @@ static int resolve_config_file(void)
     char *xdg = getenv("XDG_CONFIG_HOME");
     if (xdg)
     {
-        asprintf(&c.path, "%s/tio/config", xdg);
-        if (access(c.path, F_OK) == 0)
+        if (asprintf(&c.path, "%s/tio/config", xdg) != -1)
         {
-            return 0;
+            if (access(c.path, F_OK) == 0)
+            {
+                return 0;
+            }
+            free(c.path);
         }
-        free(c.path);
     }
 
     char *home = getenv("HOME");
     if (home)
     {
-        asprintf(&c.path, "%s/.config/tio/config", home);
-        if (access(c.path, F_OK) == 0)
+        if (asprintf(&c.path, "%s/.config/tio/config", home) != -1)
         {
-            return 0;
+            if (access(c.path, F_OK) == 0)
+            {
+                return 0;
+            }
+            free(c.path);
         }
-        free(c.path);
 
-        asprintf(&c.path, "%s/.tioconfig", home);
-        if (access(c.path, F_OK) == 0)
+        if (asprintf(&c.path, "%s/.tioconfig", home) != -1)
         {
-            return 0;
+            if (access(c.path, F_OK) == 0)
+            {
+                return 0;
+            }
+            free(c.path);
         }
-        free(c.path);
     }
 
     c.path = NULL;
