@@ -1,5 +1,5 @@
 /*
- * tio - a simple serial terminal I/O tool
+ * tio - a serial device I/O tool
  *
  * Copyright (c) 2014-2022  Martin Lund
  *
@@ -21,6 +21,9 @@
 
 #pragma once
 
+#include <stdbool.h>
+#include <stdio.h>
+
 #define UNUSED(expr) do { (void)(expr); } while (0)
 
 char * current_time(void);
@@ -29,6 +32,6 @@ long string_to_long(char *string);
 int ctrl_key_code(unsigned char key);
 void alert_connect(void);
 void alert_disconnect(void);
-
-extern char key_hit;
-int xymodem_send(int sio, const char *filename, char mode);
+bool fs_dir_exists(const char *path);
+bool regex_match(const char *string, const char *pattern);
+int read_poll(int fd, void *data, size_t len, int timeout);
