@@ -1,5 +1,5 @@
 /*
- * tio - a simple serial terminal I/O tool
+ * tio - a serial device I/O tool
  *
  * Copyright (c) 2014-2022  Martin Lund
  *
@@ -39,21 +39,27 @@
 #include "timestamp.h"
 #include "alert.h"
 #include "log.h"
+#include "script.h"
 
 enum opt_t
 {
     OPT_NONE,
     OPT_TIMESTAMP_FORMAT,
     OPT_LOG_FILE,
+    OPT_LOG_DIRECTORY,
     OPT_LOG_STRIP,
     OPT_LOG_APPEND,
     OPT_LINE_PULSE_DURATION,
-    OPT_RESPONSE_TIMEOUT,
     OPT_RS485,
     OPT_RS485_CONFIG,
     OPT_ALERT,
     OPT_COMPLETE_SUB_CONFIGS,
     OPT_MUTE,
+    OPT_SCRIPT,
+    OPT_SCRIPT_FILE,
+    OPT_SCRIPT_RUN,
+    OPT_INPUT_MODE,
+    OPT_OUTPUT_MODE,
 };
 
 /* Default options */
@@ -77,17 +83,18 @@ struct option_t option =
     .log = false,
     .log_append = false,
     .log_filename = NULL,
+    .log_directory = NULL,
     .log_strip = false,
     .local_echo = false,
     .timestamp = TIMESTAMP_NONE,
     .socket = NULL,
     .map = "",
     .color = 256, // Bold
-    .hex_mode = false,
+    .input_mode = INPUT_MODE_NORMAL,
+    .output_mode = OUTPUT_MODE_NORMAL,
     .prefix_code = 20, // ctrl-t
     .prefix_key = 't',
-    .response_wait = false,
-    .response_timeout = 100,
+    .prefix_enabled = true,
     .mute = false,
     .rs485 = false,
     .rs485_config_flags = 0,
@@ -95,6 +102,9 @@ struct option_t option =
     .rs485_delay_rts_after_send = -1,
     .alert = ALERT_NONE,
     .complete_sub_configs = false,
+    .script = NULL,
+    .script_filename = NULL,
+    .script_run = SCRIPT_RUN_ALWAYS,
 };
 
 void print_help(char *argv[])
@@ -116,23 +126,26 @@ void print_help(char *argv[])
     printf("      --line-pulse-duration <duration>   Set line pulse duration\n");
     printf("  -n, --no-autoconnect                   Disable automatic connect\n");
     printf("  -e, --local-echo                       Enable local echo\n");
+    printf("      --input-mode normal|hex            Select input mode (default: normal)\n");
+    printf("      --output-mode normal|hex           Select output mode (default: normal)\n");
     printf("  -t, --timestamp                        Enable line timestamp\n");
     printf("      --timestamp-format <format>        Set timestamp format (default: 24hour)\n");
     printf("  -L, --list-devices                     List available serial devices by ID\n");
     printf("  -l, --log                              Enable log to file\n");
     printf("      --log-file <filename>              Set log filename\n");
+    printf("      --log-directory <path>             Set log directory path for automatic named logs\n");
     printf("      --log-append                       Append to log file\n");
     printf("      --log-strip                        Strip control characters and escape sequences\n");
     printf("  -m, --map <flags>                      Map characters\n");
     printf("  -c, --color 0..255|bold|none|list      Colorize tio text (default: bold)\n");
     printf("  -S, --socket <socket>                  Redirect I/O to socket\n");
-    printf("  -x, --hexadecimal                      Enable hexadecimal mode\n");
-    printf("  -r, --response-wait                    Wait for line response then quit\n");
-    printf("      --response-timeout <ms>            Response timeout (default: 100)\n");
     printf("      --rs-485                           Enable RS-485 mode\n");
     printf("      --rs-485-config <config>           Set RS-485 configuration\n");
     printf("      --alert bell|blink|none            Alert on connect/disconnect (default: none)\n");
     printf("      --mute                             Mute tio\n");
+    printf("      --script <string>                  Run script from string\n");
+    printf("      --script-file <filename>           Run script from file\n");
+    printf("      --script-run once|always|never     Run script on connect (default: always)\n");
     printf("  -v, --version                          Display version\n");
     printf("  -h, --help                             Display help\n");
     printf("\n");
@@ -162,31 +175,37 @@ void line_pulse_duration_option_parse(const char *arg)
         {
             char keyname[11];
             unsigned int value;
-            sscanf(token, "%10[^=]=%d", keyname, &value);
 
-            if (!strcmp(keyname, "DTR"))
+            if (sscanf(token, "%10[^=]=%d", keyname, &value) == 2)
             {
-                option.dtr_pulse_duration = value;
+                if (!strcmp(keyname, "DTR"))
+                {
+                    option.dtr_pulse_duration = value;
+                }
+                else if (!strcmp(keyname, "RTS"))
+                {
+                    option.rts_pulse_duration = value;
+                }
+                else if (!strcmp(keyname, "CTS"))
+                {
+                    option.cts_pulse_duration = value;
+                }
+                else if (!strcmp(keyname, "DSR"))
+                {
+                    option.dsr_pulse_duration = value;
+                }
+                else if (!strcmp(keyname, "DCD"))
+                {
+                    option.dcd_pulse_duration = value;
+                }
+                else if (!strcmp(keyname, "RI"))
+                {
+                    option.ri_pulse_duration = value;
+                }
             }
-            else if (!strcmp(keyname, "RTS"))
+            else
             {
-                option.rts_pulse_duration = value;
-            }
-            else if (!strcmp(keyname, "CTS"))
-            {
-                option.cts_pulse_duration = value;
-            }
-            else if (!strcmp(keyname, "DSR"))
-            {
-                option.dsr_pulse_duration = value;
-            }
-            else if (!strcmp(keyname, "DCD"))
-            {
-                option.dcd_pulse_duration = value;
-            }
-            else if (!strcmp(keyname, "RI"))
-            {
-                option.ri_pulse_duration = value;
+                token_found = false;
             }
         }
         else
@@ -197,6 +216,91 @@ void line_pulse_duration_option_parse(const char *arg)
     free(buffer);
 }
 
+input_mode_t input_mode_option_parse(const char *arg)
+{
+    if (strcmp("normal", arg) == 0)
+    {
+        return INPUT_MODE_NORMAL;
+    }
+    else if (strcmp("hex", arg) == 0)
+    {
+        return INPUT_MODE_HEX;
+    }
+    else
+    {
+        tio_error_printf("Invalid input mode option");
+        exit(EXIT_FAILURE);
+    }
+}
+
+output_mode_t output_mode_option_parse(const char *arg)
+{
+    if (strcmp("normal", arg) == 0)
+    {
+        return OUTPUT_MODE_NORMAL;
+    }
+    else if (strcmp("hex", arg) == 0)
+    {
+        return OUTPUT_MODE_HEX;
+    }
+    else
+    {
+        tio_error_printf("Invalid output mode option");
+        exit(EXIT_FAILURE);
+    }
+}
+
+const char *input_mode_by_string(input_mode_t mode)
+{
+    switch (mode)
+    {
+        case INPUT_MODE_NORMAL:
+            return "normal";
+        case INPUT_MODE_HEX:
+            return "hex";
+        case INPUT_MODE_END:
+            break;
+    }
+
+    return NULL;
+}
+
+const char *output_mode_by_string(output_mode_t mode)
+{
+    switch (mode)
+    {
+        case OUTPUT_MODE_NORMAL:
+            return "normal";
+        case OUTPUT_MODE_HEX:
+            return "hex";
+        case OUTPUT_MODE_END:
+            break;
+    }
+
+    return NULL;
+}
+
+enum script_run_t script_run_option_parse(const char *arg)
+{
+    if (strcmp("once", arg) == 0)
+    {
+        return SCRIPT_RUN_ONCE;
+    }
+    else if (strcmp("always", arg) == 0)
+    {
+        return SCRIPT_RUN_ALWAYS;
+    }
+    else if (strcmp("never", arg) == 0)
+    {
+        return SCRIPT_RUN_NEVER;
+    }
+    else
+    {
+        tio_error_printf("Invalid script run option");
+        exit(EXIT_FAILURE);
+    }
+}
+
 void options_print()
 {
     tio_printf(" Device: %s", option.tty_device);
@@ -216,7 +320,8 @@ void options_print()
                                                                             option.dsr_pulse_duration,
                                                                             option.dcd_pulse_duration,
                                                                             option.ri_pulse_duration);
-    tio_printf(" Hexadecimal mode: %s", option.hex_mode ? "enabled" : "disabled");
+    tio_printf(" Input mode: %s", input_mode_by_string(option.input_mode));
+    tio_printf(" Output mode: %s", output_mode_by_string(option.output_mode));
     if (option.map[0] != 0)
         tio_printf(" Map flags: %s", option.map);
     if (option.log)
@@ -235,6 +340,11 @@ void options_parse(int argc, char *argv[])
         exit(EXIT_SUCCESS);
     }
 
+    // Support no-color.org informal spec
+    char *no_color = getenv("NO_COLOR");
+    if(no_color != NULL && no_color[0] != '\0')
+	    option.color = -1;
+
     while (1)
     {
         static struct option long_options[] =
@@ -254,18 +364,21 @@ void options_parse(int argc, char *argv[])
             {"list-devices",         no_argument,       0, 'L'                     },
             {"log",                  no_argument,       0, 'l'                     },
             {"log-file",             required_argument, 0, OPT_LOG_FILE            },
+            {"log-directory",        required_argument, 0, OPT_LOG_DIRECTORY       },
             {"log-append",           no_argument,       0, OPT_LOG_APPEND          },
             {"log-strip",            no_argument,       0, OPT_LOG_STRIP           },
             {"socket",               required_argument, 0, 'S'                     },
             {"map",                  required_argument, 0, 'm'                     },
             {"color",                required_argument, 0, 'c'                     },
-            {"hexadecimal",          no_argument,       0, 'x'                     },
-            {"response-wait",        no_argument,       0, 'r'                     },
-            {"response-timeout",     required_argument, 0, OPT_RESPONSE_TIMEOUT    },
+            {"input-mode",           required_argument, 0, OPT_INPUT_MODE          },
+            {"output-mode",          required_argument, 0, OPT_OUTPUT_MODE         },
             {"rs-485",               no_argument,       0, OPT_RS485               },
             {"rs-485-config",        required_argument, 0, OPT_RS485_CONFIG        },
             {"alert",                required_argument, 0, OPT_ALERT               },
             {"mute",                 no_argument,       0, OPT_MUTE                },
+            {"script",               required_argument, 0, OPT_SCRIPT              },
+            {"script-file",          required_argument, 0, OPT_SCRIPT_FILE         },
+            {"script-run",           required_argument, 0, OPT_SCRIPT_RUN          },
             {"version",              no_argument,       0, 'v'                     },
             {"help",                 no_argument,       0, 'h'                     },
             {"complete-sub-configs", no_argument,       0, OPT_COMPLETE_SUB_CONFIGS},
@@ -355,6 +468,10 @@ void options_parse(int argc, char *argv[])
                 option.log_filename = optarg;
                 break;
 
+            case OPT_LOG_DIRECTORY:
+                option.log_directory = optarg;
+                break;
+
             case OPT_LOG_STRIP:
                 option.log_strip = true;
                 break;
@@ -401,16 +518,12 @@ void options_parse(int argc, char *argv[])
                 }
                 break;
 
-            case 'x':
-                option.hex_mode = true;
+            case OPT_INPUT_MODE:
+                option.input_mode = input_mode_option_parse(optarg);
                 break;
 
-            case 'r':
-                option.response_wait = true;
-                break;
-
-            case OPT_RESPONSE_TIMEOUT:
-                option.response_timeout = string_to_long(optarg);
+            case OPT_OUTPUT_MODE:
+                option.output_mode = output_mode_option_parse(optarg);
                 break;
 
             case OPT_RS485:
@@ -429,6 +542,18 @@ void options_parse(int argc, char *argv[])
                 option.mute = true;
                 break;
 
+            case OPT_SCRIPT:
+                option.script = optarg;
+                break;
+
+            case OPT_SCRIPT_FILE:
+                option.script_filename = optarg;
+                break;
+
+            case OPT_SCRIPT_RUN:
+                option.script_run = script_run_option_parse(optarg);
+                break;
+
             case 'v':
                 printf("tio v%s\n", VERSION);
                 exit(EXIT_SUCCESS);
