@@ -1,5 +1,5 @@
 /*
- * tio - a simple serial terminal I/O tool
+ * tio - a serial device I/O tool
  *
  * Copyright (c) 2014-2022  Martin Lund
  *
@@ -55,6 +55,8 @@
 #include "alert.h"
 #include "timestamp.h"
 #include "misc.h"
+#include "script.h"
+#include "xymodem.h"
 
 /* tty device listing configuration */
 
@@ -97,12 +99,14 @@
 #define KEY_F 0x66
 #define KEY_SHIFT_F 0x46
 #define KEY_G 0x67
-#define KEY_H 0x68
+#define KEY_I 0x69
 #define KEY_L 0x6C
 #define KEY_SHIFT_L 0x4C
 #define KEY_M 0x6D
+#define KEY_O 0x6F
 #define KEY_P 0x70
 #define KEY_Q 0x71
+#define KEY_R 0x72
 #define KEY_S 0x73
 #define KEY_T 0x74
 #define KEY_U 0x55
@@ -111,12 +115,26 @@
 #define KEY_Y 0x79
 #define KEY_Z 0x7a
 
-enum line_mode_t
+typedef enum
 {
-    LINE_OFF,
     LINE_TOGGLE,
     LINE_PULSE
-};
+} tty_line_mode_t;
+
+typedef enum
+{
+    SUBCOMMAND_NONE,
+    SUBCOMMAND_LINE_TOGGLE,
+    SUBCOMMAND_LINE_PULSE,
+    SUBCOMMAND_XMODEM,
+} sub_command_t;
+
+typedef struct
+{
+    int mask;
+    bool value;
+    bool reserved;
+} tty_line_config_t;
 
 const char random_array[] =
 {
@@ -144,12 +162,14 @@ static unsigned long rx_total = 0, tx_total = 0;
 static bool connected = false;
 static bool standard_baudrate = true;
 static void (*print)(char c);
-static int fd;
+static int device_fd;
+static bool map_i_ff_escc = false;
 static bool map_i_nl_crnl = false;
 static bool map_o_cr_nl = false;
 static bool map_o_nl_crnl = false;
 static bool map_o_del_bs = false;
 static bool map_o_ltu = false;
+static bool map_o_nulbrk = false;
 static bool map_o_msblsb = false;
 static char hex_chars[2];
 static unsigned char hex_char_index = 0;
@@ -160,6 +180,7 @@ static pthread_t thread;
 static int pipefd[2];
 static pthread_mutex_t mutex_input_ready = PTHREAD_MUTEX_INITIALIZER;
 static char line[LINE_SIZE_MAX];
+static tty_line_config_t line_config[6] = { };
 
 static void optional_local_echo(char c)
 {
@@ -167,11 +188,15 @@ static void optional_local_echo(char c)
     {
         return;
     }
+
     print(c);
+
     if (option.log)
     {
         log_putc(c);
     }
+
+    print_tainted_set();
 }
 
 inline static bool is_valid_hex(char c)
@@ -338,7 +363,7 @@ void *tty_stdin_input_thread(void *arg)
 
                 input_char = input_buffer[i];
 
-                if (previous_char == option.prefix_code)
+                if (option.prefix_enabled && previous_char == option.prefix_code)
                 {
                     if (input_char == option.prefix_code)
                     {
@@ -353,7 +378,7 @@ void *tty_stdin_input_thread(void *arg)
                             break;
                         case KEY_SHIFT_F:
                             tio_printf("Flushed data I/O channels")
-                            tcflush(fd, TCIOFLUSH);
+                            tcflush(device_fd, TCIOFLUSH);
                             break;
                         default:
                             break;
@@ -394,24 +419,30 @@ void tty_input_thread_wait_ready(void)
     pthread_mutex_lock(&mutex_input_ready);
 }
 
-static void output_hex(char c)
+static void handle_hex_prompt(char c)
 {
     hex_chars[hex_char_index++] = c;
 
     printf("%c", c);
+    print_tainted_set();
 
     if (hex_char_index == 2)
     {
         usleep(100*1000);
-        printf("\b \b");
-        printf("\b \b");
+        if (option.local_echo == false)
+        {
+            printf("\b \b");
+            printf("\b \b");
+        }
+        else
+        {
+            printf(" ");
+        }
 
         unsigned char hex_value = char_to_nibble(hex_chars[0]) << 4 | (char_to_nibble(hex_chars[1]) & 0x0F);
         hex_char_index = 0;
 
-        optional_local_echo(hex_value);
-
-        ssize_t status = tty_write(fd, &hex_value, 1);
+        ssize_t status = tty_write(device_fd, &hex_value, 1);
         if (status < 0)
         {
             tio_warning_printf("Could not write to tty device");
@@ -423,67 +454,166 @@ static void output_hex(char c)
     }
 }
 
-static void toggle_line(const char *line_name, int mask, enum line_mode_t line_mode)
+static const char *tty_line_name(int mask)
 {
-    int state;
-
-    if (line_mode == LINE_TOGGLE)
+    switch (mask)
     {
-        // Toggle line
-        if (ioctl(fd, TIOCMGET, &state) < 0)
+        case TIOCM_DTR:
+            return "DTR";
+        case TIOCM_RTS:
+            return "RTS";
+        case TIOCM_CTS:
+            return "CTS";
+        case TIOCM_DSR:
+            return "DSR";
+        case TIOCM_CD:
+            return "CD";
+        case TIOCM_RI:
+            return "RI";
+        default:
+            return NULL;
+    }
+}
+
+void tty_line_config(int mask, bool value)
+{
+    int i = 0;
+
+    for (i=0; i<6; i++)
+    {
+        if ((line_config[i].mask == mask) || (line_config[i].reserved == false))
         {
-            tio_warning_printf("Could not get line state (%s)", strerror(errno));
+            line_config[i].mask = mask;
+            line_config[i].value = value;
+            line_config[i].reserved = true;
+            break;
         }
-        else
+    }
+}
+
+void tty_line_config_apply(void)
+{
+    int i = 0;
+    static int state;
+
+    if (ioctl(device_fd, TIOCMGET, &state) < 0)
+    {
+        tio_warning_printf("Could not get line state (%s)", strerror(errno));
+        return;
+    }
+
+    for (i=0; i<6; i++)
+    {
+        if (line_config[i].reserved)
         {
-            if (state & mask)
+            if (line_config[i].value)
             {
-                state &= ~mask;
-                tio_printf("Setting %s to LOW", line_name);
+                // High
+                state &= ~line_config[i].mask;
+                tio_printf("Setting %s to HIGH", tty_line_name(line_config[i].mask));
             }
             else
             {
-                state |= mask;
-                tio_printf("Setting %s to HIGH", line_name);
+                // Low
+                state |= line_config[i].mask;
+                tio_printf("Setting %s to LOW", tty_line_name(line_config[i].mask));
             }
-            if (ioctl(fd, TIOCMSET, &state) < 0)
-                tio_warning_printf("Could not set line state (%s)", strerror(errno));
+
+            line_config[i].reserved = true;
         }
-    } else if (line_mode == LINE_PULSE)
+    }
+
+    if (ioctl(device_fd, TIOCMSET, &state) < 0)
     {
-        int duration = 0;
-        // Pulse line
-        toggle_line(line_name, mask, LINE_TOGGLE);
-        switch (mask)
-        {
-            case TIOCM_DTR:
-                duration = option.dtr_pulse_duration;
-                break;
-            case TIOCM_RTS:
-                duration = option.rts_pulse_duration;
-                break;
-            case TIOCM_CTS:
-                duration = option.cts_pulse_duration;
-                break;
-            case TIOCM_DSR:
-                duration = option.dsr_pulse_duration;
-                break;
-            case TIOCM_CD:
-                duration = option.dcd_pulse_duration;
-                break;
-            case TIOCM_RI:
-                duration = option.ri_pulse_duration;
-                break;
-            default:
-                duration = 0;
-                break;
-        }
-        if (duration > 0)
-        {
-            tio_printf("Waiting %d ms", duration);
-            delay(duration);
-        }
-        toggle_line(line_name, mask, LINE_TOGGLE);
+        tio_warning_printf("Could not set line state configuration (%s)", strerror(errno));
+    }
+
+    // Reset configuration
+    for (i=0; i<6; i++)
+    {
+        line_config[i].reserved = false;
+        line_config[i].mask = -1;
+    }
+}
+
+void tty_line_set(int fd, int mask, bool value)
+{
+    int state;
+
+    if (ioctl(fd, TIOCMGET, &state) < 0)
+    {
+        tio_warning_printf("Could not get line state (%s)", strerror(errno));
+        return;
+    }
+
+    if (value)
+    {
+        state &= ~mask;
+        tio_printf("Setting %s to HIGH", tty_line_name(mask));
+    }
+    else
+    {
+        state |= mask;
+        tio_printf("Setting %s to LOW", tty_line_name(mask));
+    }
+
+    if (ioctl(fd, TIOCMSET, &state) < 0)
+    {
+        tio_warning_printf("Could not set line state (%s)", strerror(errno));
+    }
+}
+
+void tty_line_toggle(int fd, int mask)
+{
+    int state;
+
+    if (ioctl(fd, TIOCMGET, &state) < 0)
+    {
+        tio_warning_printf("Could not get line state (%s)", strerror(errno));
+        return;
+    }
+
+    if (state & mask)
+    {
+        state &= ~mask;
+        tio_printf("Setting %s to HIGH", tty_line_name(mask));
+    }
+    else
+    {
+        state |= mask;
+        tio_printf("Setting %s to LOW", tty_line_name(mask));
+    }
+
+    if (ioctl(fd, TIOCMSET, &state) < 0)
+    {
+        tio_warning_printf("Could not set line state (%s)", strerror(errno));
+    }
+}
+
+static void tty_line_pulse(int fd, int mask, unsigned int duration)
+{
+    tty_line_toggle(fd, mask);
+
+    if (duration > 0)
+    {
+        tio_printf("Waiting %d ms", duration);
+        delay(duration);
+    }
+
+    tty_line_toggle(fd, mask);
+}
+
+static void tty_line_poke(int fd, int mask, tty_line_mode_t mode, unsigned int duration)
+{
+    switch (mode)
+    {
+        case LINE_TOGGLE:
+            tty_line_toggle(fd, mask);
+            break;
+
+        case LINE_PULSE:
+            tty_line_pulse(fd, mask, duration);
+            break;
     }
 }
 
@@ -514,12 +644,30 @@ static int tio_readln(void)
     return (p - line);
 }
 
+void tty_output_mode_set(output_mode_t mode)
+{
+    switch (mode)
+    {
+        case OUTPUT_MODE_NORMAL:
+            print = print_normal;
+            break;
+
+        case OUTPUT_MODE_HEX:
+            print = print_hex;
+            break;
+
+        case OUTPUT_MODE_END:
+            break;
+    }
+}
+
 void handle_command_sequence(char input_char, char *output_char, bool *forward)
 {
     char unused_char;
     bool unused_bool;
     int state;
-    static enum line_mode_t line_mode = LINE_OFF;
+    static tty_line_mode_t line_mode;
+    static sub_command_t sub_command = SUBCOMMAND_NONE;
     static char previous_char = 0;
 
     /* Ignore unused arguments */
@@ -533,42 +681,78 @@ void handle_command_sequence(char input_char, char *output_char, bool *forward)
         forward = &unused_bool;
     }
 
-    if (line_mode)
+    // Handle sub commands
+    if (sub_command)
     {
-        // Handle line toggle number action
         *forward = false;
-        switch (input_char)
+
+        switch (sub_command)
         {
-            case KEY_0:
-                toggle_line("DTR", TIOCM_DTR, line_mode);
+            case SUBCOMMAND_NONE:
                 break;
-            case KEY_1:
-                toggle_line("RTS", TIOCM_RTS, line_mode);
+
+            case SUBCOMMAND_LINE_TOGGLE:
+            case SUBCOMMAND_LINE_PULSE:
+                switch (input_char)
+                {
+                    case KEY_0:
+                        tty_line_poke(device_fd, TIOCM_DTR, line_mode, option.dtr_pulse_duration);
+                        break;
+                    case KEY_1:
+                        tty_line_poke(device_fd, TIOCM_RTS, line_mode, option.rts_pulse_duration);
+                        break;
+                    case KEY_2:
+                        tty_line_poke(device_fd, TIOCM_CTS, line_mode, option.cts_pulse_duration);
+                        break;
+                    case KEY_3:
+                        tty_line_poke(device_fd, TIOCM_DSR, line_mode, option.dsr_pulse_duration);
+                        break;
+                    case KEY_4:
+                        tty_line_poke(device_fd, TIOCM_CD, line_mode, option.dcd_pulse_duration);
+                        break;
+                    case KEY_5:
+                        tty_line_poke(device_fd, TIOCM_RI, line_mode, option.ri_pulse_duration);
+                        break;
+                    default:
+                        tio_warning_printf("Invalid line number");
+                        break;
+                }
                 break;
-            case KEY_2:
-                toggle_line("CTS", TIOCM_CTS, line_mode);
-                break;
-            case KEY_3:
-                toggle_line("DSR", TIOCM_DSR, line_mode);
-                break;
-            case KEY_4:
-                toggle_line("DCD", TIOCM_CD, line_mode);
-                break;
-            case KEY_5:
-                toggle_line("RI", TIOCM_RI, line_mode);
-                break;
-            default:
-                tio_warning_printf("Invalid line number");
+
+            case SUBCOMMAND_XMODEM:
+                switch (input_char)
+                {
+                    case KEY_0:
+                        tio_printf("Send file with XMODEM-1K");
+                        tio_printf_raw("Enter file name: ");
+                        if (tio_readln())
+                        {
+                            tio_printf("Sending file '%s'  ", line);
+                            tio_printf("Press any key to abort transfer");
+                            tio_printf("%s", xymodem_send(device_fd, line, XMODEM_CRC) < 0 ? "Aborted" : "Done");
+                        }
+                        break;
+
+                    case KEY_1:
+                        tio_printf("Send file with XMODEM-CRC");
+                        tio_printf_raw("Enter file name: ");
+                        if (tio_readln())
+                        {
+                            tio_printf("Sending file '%s'  ", line);
+                            tio_printf("Press any key to abort transfer");
+                            tio_printf("%s", xymodem_send(device_fd, line, XMODEM_CRC) < 0 ? "Aborted" : "Done");
+                        }
+                        break;
+                }
                 break;
         }
 
-        line_mode = LINE_OFF;
-
+        sub_command = SUBCOMMAND_NONE;
         return;
     }
 
     /* Handle escape key commands */
-    if (previous_char == option.prefix_code)
+    if (option.prefix_enabled && previous_char == option.prefix_code)
     {
         /* Do not forward input char to output by default */
         *forward = false;
@@ -583,6 +767,7 @@ void handle_command_sequence(char input_char, char *output_char, bool *forward)
             return;
         }
 
+        // Handle commands
         switch (input_char)
         {
             case KEY_QUESTION:
@@ -594,34 +779,36 @@ void handle_command_sequence(char input_char, char *output_char, bool *forward)
                 tio_printf(" ctrl-%c f       Toggle log to file", option.prefix_key);
                 tio_printf(" ctrl-%c F       Flush data I/O buffers", option.prefix_key);
                 tio_printf(" ctrl-%c g       Toggle serial port line", option.prefix_key);
-                tio_printf(" ctrl-%c h       Toggle hexadecimal mode", option.prefix_key);
+                tio_printf(" ctrl-%c i       Toggle input mode", option.prefix_key);
                 tio_printf(" ctrl-%c l       Clear screen", option.prefix_key);
                 tio_printf(" ctrl-%c L       Show line states", option.prefix_key);
                 tio_printf(" ctrl-%c m       Toggle MSB to LSB bit order", option.prefix_key);
+                tio_printf(" ctrl-%c o       Toggle output mode", option.prefix_key);
                 tio_printf(" ctrl-%c p       Pulse serial port line", option.prefix_key);
                 tio_printf(" ctrl-%c q       Quit", option.prefix_key);
+                tio_printf(" ctrl-%c r       Run script", option.prefix_key);
                 tio_printf(" ctrl-%c s       Show statistics", option.prefix_key);
                 tio_printf(" ctrl-%c t       Toggle line timestamp mode", option.prefix_key);
                 tio_printf(" ctrl-%c U       Toggle conversion to uppercase on output", option.prefix_key);
                 tio_printf(" ctrl-%c v       Show version", option.prefix_key);
-                tio_printf(" ctrl-%c x       Send file via Xmodem-1K", option.prefix_key);
+                tio_printf(" ctrl-%c x       Send file via Xmodem", option.prefix_key);
                 tio_printf(" ctrl-%c y       Send file via Ymodem", option.prefix_key);
-                tio_printf(" ctrl-%c ctrl-%c Send ctrl-%c character", option.prefix_key, option.prefix_key, option.prefix_key);
+                tio_printf(" ctrl-%c ctrl-%c  Send ctrl-%c character", option.prefix_key, option.prefix_key, option.prefix_key);
                 break;
 
             case KEY_SHIFT_L:
-                if (ioctl(fd, TIOCMGET, &state) < 0)
+                if (ioctl(device_fd, TIOCMGET, &state) < 0)
                 {
                     tio_warning_printf("Could not get line state (%s)", strerror(errno));
                     break;
                 }
                 tio_printf("Line states:");
-                tio_printf(" DTR: %s", (state & TIOCM_DTR) ? "HIGH" : "LOW");
-                tio_printf(" RTS: %s", (state & TIOCM_RTS) ? "HIGH" : "LOW");
-                tio_printf(" CTS: %s", (state & TIOCM_CTS) ? "HIGH" : "LOW");
-                tio_printf(" DSR: %s", (state & TIOCM_DSR) ? "HIGH" : "LOW");
-                tio_printf(" DCD: %s", (state & TIOCM_CD) ? "HIGH" : "LOW");
-                tio_printf(" RI : %s", (state & TIOCM_RI) ? "HIGH" : "LOW");
+                tio_printf(" DTR: %s", (state & TIOCM_DTR) ? "LOW" : "HIGH");
+                tio_printf(" RTS: %s", (state & TIOCM_RTS) ? "LOW" : "HIGH");
+                tio_printf(" CTS: %s", (state & TIOCM_CTS) ? "LOW" : "HIGH");
+                tio_printf(" DSR: %s", (state & TIOCM_DSR) ? "LOW" : "HIGH");
+                tio_printf(" DCD: %s", (state & TIOCM_CD) ? "LOW" : "HIGH");
+                tio_printf(" RI : %s", (state & TIOCM_RI) ? "LOW" : "HIGH");
                 break;
 
             case KEY_F:
@@ -645,30 +832,32 @@ void handle_command_sequence(char input_char, char *output_char, bool *forward)
 
             case KEY_G:
                 tio_printf("Please enter which serial line number to toggle:");
-                tio_printf(" DTR (0)");
-                tio_printf(" RTS (1)");
-                tio_printf(" CTS (2)");
-                tio_printf(" DSR (3)");
-                tio_printf(" DCD (4)");
-                tio_printf(" RI  (5)");
-                // Process next input character as part of the line toggle step
+                tio_printf("(0) DTR");
+                tio_printf("(1) RTS");
+                tio_printf("(2) CTS");
+                tio_printf("(3) DSR");
+                tio_printf("(4) DCD");
+                tio_printf("(5) RI");
                 line_mode = LINE_TOGGLE;
+                // Process next input character as sub command
+                sub_command = SUBCOMMAND_LINE_TOGGLE;
                 break;
 
             case KEY_P:
                 tio_printf("Please enter which serial line number to pulse:");
-                tio_printf(" DTR (0)");
-                tio_printf(" RTS (1)");
-                tio_printf(" CTS (2)");
-                tio_printf(" DSR (3)");
-                tio_printf(" DCD (4)");
-                tio_printf(" RI  (5)");
-                // Process next input character as part of the line pulse step
+                tio_printf("(0) DTR");
+                tio_printf("(1) RTS");
+                tio_printf("(2) CTS");
+                tio_printf("(3) DSR");
+                tio_printf("(4) DCD");
+                tio_printf("(5) RI");
                 line_mode = LINE_PULSE;
+                // Process next input character as sub command
+                sub_command = SUBCOMMAND_LINE_PULSE;
                 break;
 
             case KEY_B:
-                tcsendbreak(fd, 0);
+                tcsendbreak(device_fd, 0);
                 break;
 
             case KEY_C:
@@ -686,19 +875,42 @@ void handle_command_sequence(char input_char, char *output_char, bool *forward)
                 tio_printf("Switched local echo %s", option.local_echo ? "on" : "off");
                 break;
 
-            case KEY_H:
-                /* Toggle hexadecimal printing mode */
-                if (!option.hex_mode)
+            case KEY_I:
+                option.input_mode += 1;
+                switch (option.input_mode)
                 {
-                    print = print_hex;
-                    option.hex_mode = true;
-                    tio_printf("Switched to hexadecimal mode");
+                    case INPUT_MODE_NORMAL:
+                        break;
+
+                    case INPUT_MODE_HEX:
+                        option.input_mode = INPUT_MODE_HEX;
+                        tio_printf("Switched to hex input mode");
+                        break;
+
+                    case INPUT_MODE_END:
+                        option.input_mode = INPUT_MODE_NORMAL;
+                        tio_printf("Switched to normal input mode");
+                        break;
                 }
-                else
+                break;
+
+            case KEY_O:
+                option.output_mode += 1;
+                switch (option.output_mode)
                 {
-                    print = print_normal;
-                    option.hex_mode = false;
-                    tio_printf("Switched to normal mode");
+                    case OUTPUT_MODE_NORMAL:
+                        break;
+
+                    case OUTPUT_MODE_HEX:
+                        tty_output_mode_set(OUTPUT_MODE_HEX);
+                        tio_printf("Switched to hex output mode");
+                        break;
+
+                    case OUTPUT_MODE_END:
+                        option.output_mode = OUTPUT_MODE_NORMAL;
+                        tty_output_mode_set(OUTPUT_MODE_NORMAL);
+                        tio_printf("Switched to normal output mode");
+                        break;
                 }
                 break;
 
@@ -725,6 +937,11 @@ void handle_command_sequence(char input_char, char *output_char, bool *forward)
                 /* Exit upon ctrl-t q sequence */
                 exit(EXIT_SUCCESS);
 
+            case KEY_R:
+                /* Run script */
+                script_run(device_fd);
+                break;
+
             case KEY_S:
                 /* Show tx/rx statistics upon ctrl-t s sequence */
                 tio_printf("Statistics:");
@@ -766,13 +983,20 @@ void handle_command_sequence(char input_char, char *output_char, bool *forward)
                 break;
 
             case KEY_X:
+                tio_printf("Please enter which X modem protocol to use:");
+                tio_printf(" (0) XMODEM-1K");
+                tio_printf(" (1) XMODEM-CRC");
+                // Process next input character as sub command
+                sub_command = SUBCOMMAND_XMODEM;
+                break;
+
             case KEY_Y:
-                tio_printf("Send file with %cMODEM", toupper(input_char));
+                tio_printf("Send file with YMODEM");
                 tio_printf_raw("Enter file name: ");
                 if (tio_readln()) {
                     tio_printf("Sending file '%s'  ", line);
                     tio_printf("Press any key to abort transfer");
-                    tio_printf("%s", xymodem_send(fd, line, input_char) < 0 ? "Aborted" : "Done");
+                    tio_printf("%s", xymodem_send(device_fd, line, YMODEM) < 0 ? "Aborted" : "Done");
                 }
                 break;
 
@@ -1070,6 +1294,10 @@ void tty_configure(void)
             {
                 map_o_del_bs = true;
             }
+            else if (strcmp(token,"IFFESCC") == 0)
+            {
+                map_i_ff_escc = true;
+            }
             else if (strcmp(token,"INLCRNL") == 0)
             {
                 map_i_nl_crnl = true;
@@ -1082,6 +1310,10 @@ void tty_configure(void)
             {
                 map_o_ltu = true;
             }
+            else if (strcmp(token, "ONULBRK") == 0)
+            {
+                map_o_nulbrk = true;
+            }
             else if (strcmp(token, "MSB2LSB") == 0)
             {
                 map_o_msblsb = true;
@@ -1192,8 +1424,8 @@ void tty_disconnect(void)
     if (connected)
     {
         tio_printf("Disconnected");
-        flock(fd, LOCK_UN);
-        close(fd);
+        flock(device_fd, LOCK_UN);
+        close(device_fd);
         connected = false;
 
         /* Fire alert action */
@@ -1203,12 +1435,12 @@ void tty_disconnect(void)
 
 void tty_restore(void)
 {
-    tcsetattr(fd, TCSANOW, &tio_old);
+    tcsetattr(device_fd, TCSANOW, &tio_old);
 
     if (option.rs485)
     {
         /* Restore original RS-485 mode */
-        rs485_mode_restore(fd);
+        rs485_mode_restore(device_fd);
     }
 
     if (connected)
@@ -1248,22 +1480,48 @@ void forward_to_tty(int fd, char output_char)
     }
     else
     {
-        if (option.hex_mode)
+        switch (option.output_mode)
         {
-            output_hex(output_char);
-        }
-        else
-        {
-            /* Send output to tty device */
-            optional_local_echo(output_char);
-            status = tty_write(fd, &output_char, 1);
-            if (status < 0)
-            {
-                tio_warning_printf("Could not write to tty device");
-            }
+            case OUTPUT_MODE_NORMAL:
+                if (option.input_mode == INPUT_MODE_HEX)
+                {
+                    handle_hex_prompt(output_char);
+                }
+                else
+                {
+                    /* Send output to tty device */
+                    optional_local_echo(output_char);
+                    if ((output_char == 0) && (map_o_nulbrk))
+                    {
+                        status = tcsendbreak(fd, 0);
+                    }
+                    else
+                    {
+                        status = tty_write(fd, &output_char, 1);
+                    }
+                    if (status < 0)
+                    {
+                        tio_warning_printf("Could not write to tty device");
+                    }
 
-            /* Update transmit statistics */
-            tx_total++;
+                    /* Update transmit statistics */
+                    tx_total++;
+                }
+                break;
+
+            case OUTPUT_MODE_HEX:
+                if (option.input_mode == INPUT_MODE_HEX)
+                {
+                    handle_hex_prompt(output_char);
+                }
+                else
+                {
+                    optional_local_echo(output_char);
+                }
+                break;
+
+            case OUTPUT_MODE_END:
+                break;
         }
     }
 }
@@ -1278,27 +1536,24 @@ int tty_connect(void)
     int    status;
     bool   next_timestamp = false;
     char*  now = NULL;
-    struct timeval tv;
-    struct timeval *tv_p = &tv;
-    bool ignore_stdin = false;
 
     /* Open tty device */
-    fd = open(option.tty_device, O_RDWR | O_NOCTTY | O_NONBLOCK);
-    if (fd < 0)
+    device_fd = open(option.tty_device, O_RDWR | O_NOCTTY | O_NONBLOCK);
+    if (device_fd < 0)
     {
         tio_error_printf_silent("Could not open tty device (%s)", strerror(errno));
         goto error_open;
     }
 
     /* Make sure device is of tty type */
-    if (!isatty(fd))
+    if (!isatty(device_fd))
     {
         tio_error_printf("Not a tty device");
         exit(EXIT_FAILURE);;
     }
 
     /* Lock device file */
-    status = flock(fd, LOCK_EX | LOCK_NB);
+    status = flock(device_fd, LOCK_EX | LOCK_NB);
     if ((status == -1) && (errno == EWOULDBLOCK))
     {
         tio_error_printf("Device file is locked by another process");
@@ -1306,7 +1561,7 @@ int tty_connect(void)
     }
 
     /* Flush stale I/O data (if any) */
-    tcflush(fd, TCIOFLUSH);
+    tcflush(device_fd, TCIOFLUSH);
 
     /* Print connect status */
     tio_printf("Connected");
@@ -1322,17 +1577,10 @@ int tty_connect(void)
     }
 
     /* Manage print output mode */
-    if (option.hex_mode)
-    {
-        print = print_hex;
-    }
-    else
-    {
-        print = print_normal;
-    }
+    tty_output_mode_set(option.output_mode);
 
     /* Save current port settings */
-    if (tcgetattr(fd, &tio_old) < 0)
+    if (tcgetattr(device_fd, &tio_old) < 0)
     {
         tio_error_printf_silent("Could not get port settings (%s)", strerror(errno));
         goto error_tcgetattr;
@@ -1350,7 +1598,7 @@ int tty_connect(void)
     /* Manage RS-485 mode */
     if (option.rs485)
     {
-        rs485_mode_enable(fd);
+        rs485_mode_enable(device_fd);
     }
 
     /* Make sure we restore tty settings on exit */
@@ -1361,7 +1609,7 @@ int tty_connect(void)
     }
 
     /* Activate new port settings */
-    status = tcsetattr(fd, TCSANOW, &tio);
+    status = tcsetattr(device_fd, TCSANOW, &tio);
     if (status == -1)
     {
         tio_error_printf_silent("Could not apply port settings (%s)", strerror(errno));
@@ -1371,47 +1619,78 @@ int tty_connect(void)
     /* Set arbitrary baudrate (only works on supported platforms) */
     if (!standard_baudrate)
     {
-        if (setspeed(fd, option.baudrate) != 0)
+        if (setspeed(device_fd, option.baudrate) != 0)
         {
             tio_error_printf_silent("Could not set baudrate speed (%s)", strerror(errno));
             goto error_setspeed;
         }
     }
 
+    /* If stdin is a pipe forward all input to tty device */
+    if (interactive_mode == false)
+    {
+        while (true)
+        {
+            int ret = read(pipefd[0], &input_char, 1);
+            if (ret < 0)
+            {
+                tio_error_printf("Could not read from pipe (%s)", strerror(errno));
+                exit(EXIT_FAILURE);
+            }
+            else if (ret > 0)
+            {
+                // Forward to tty device
+                ret = write(device_fd, &input_char, 1);
+                if (ret < 0)
+                {
+                    tio_error_printf("Could not write to serial device (%s)", strerror(errno));
+                    exit(EXIT_FAILURE);
+                }
+            }
+            else
+            {
+                // EOF - finished forwarding
+                break;
+            }
+        }
+    }
+
+    /* Manage script activation */
+    if (option.script_run != SCRIPT_RUN_NEVER)
+    {
+        script_run(device_fd);
+
+        if (option.script_run == SCRIPT_RUN_ONCE)
+        {
+            option.script_run = SCRIPT_RUN_NEVER;
+        }
+    }
+
+    // Exit if piped input
+    if (interactive_mode == false)
+    {
+        exit(EXIT_SUCCESS);
+    }
+
     /* Input loop */
     while (true)
     {
         FD_ZERO(&rdfs);
-        FD_SET(fd, &rdfs);
-        if (!ignore_stdin)
-        {
-            FD_SET(pipefd[0], &rdfs);
-        }
-        maxfd = MAX(fd, pipefd[0]);
+        FD_SET(device_fd, &rdfs);
+        FD_SET(pipefd[0], &rdfs);
+
+        maxfd = MAX(device_fd, pipefd[0]);
         maxfd = MAX(maxfd, socket_add_fds(&rdfs, true));
 
-        /* Manage timeout */
-        if ((option.response_wait) && (option.response_timeout != 0))
-        {
-            // Set response timeout
-            tv_p->tv_sec = 0;
-            tv_p->tv_usec = option.response_timeout * 1000;
-        }
-        else
-        {
-            // No timeout
-            tv_p = NULL;
-        }
-
         /* Block until input becomes available */
-        status = select(maxfd + 1, &rdfs, NULL, NULL, tv_p);
+        status = select(maxfd + 1, &rdfs, NULL, NULL, NULL);
         if (status > 0)
         {
             bool forward = false;
-            if (FD_ISSET(fd, &rdfs))
+            if (FD_ISSET(device_fd, &rdfs))
             {
                 /* Input from tty device ready */
-                ssize_t bytes_read = read(fd, input_buffer, BUFSIZ);
+                ssize_t bytes_read = read(device_fd, input_buffer, BUFSIZ);
                 if (bytes_read <= 0)
                 {
                     /* Error reading - device is likely unplugged */
@@ -1428,7 +1707,7 @@ int tty_connect(void)
                     input_char = input_buffer[i];
 
                     /* Print timestamp on new line if enabled */
-                    if ((next_timestamp && input_char != '\n' && input_char != '\r') && !option.hex_mode)
+                    if ((next_timestamp && input_char != '\n' && input_char != '\r') && (option.output_mode == OUTPUT_MODE_NORMAL))
                     {
                         now = timestamp_current_time();
                         if (now)
@@ -1463,6 +1742,11 @@ int tty_connect(void)
                             next_timestamp = true;
                         }
                     }
+                    else if ((input_char == '\f') && (map_i_ff_escc) && (!map_o_msblsb))
+                    {
+                        print('\e');
+                        print('c');
+                    }
                     else
                     {
                         /* Print received tty character to stdout */
@@ -1483,15 +1767,6 @@ int tty_connect(void)
                     {
                         next_timestamp = true;
                     }
-
-                    if (option.response_wait)
-                    {
-                        if (input_char == '\n')
-                        {
-                             tty_sync(fd);
-                             exit(EXIT_SUCCESS);
-                        }
-                    }
                 }
             }
             else if (FD_ISSET(pipefd[0], &rdfs))
@@ -1505,22 +1780,9 @@ int tty_connect(void)
                 }
                 else if (bytes_read == 0)
                 {
-                    /* Reached EOF (when piping to stdin) */
-                    if (option.response_wait)
-                    {
-                        /* Stdin pipe closed but not blocking so stop listening
-                         * to stdin in response mode.
-                         *
-                         * Note: select() really indicates not if data is ready
-                         * but if file descriptor is non-blocking for I/O
-                         * operation. */
-                        ignore_stdin = true;
-                    }
-                    else
-                    {
-                        tty_sync(fd);
-                        exit(EXIT_SUCCESS);
-                    }
+                    /* Reached EOF (when piping to stdin, never reached) */
+                    tty_sync(device_fd);
+                    exit(EXIT_SUCCESS);
                 }
 
                 /* Process input byte by byte */
@@ -1535,7 +1797,7 @@ int tty_connect(void)
                     if (interactive_mode)
                     {
                         /* Do not forward prefix key */
-                        if (input_char == option.prefix_code)
+                        if (option.prefix_enabled && input_char == option.prefix_code)
                         {
                             forward = false;
                         }
@@ -1543,7 +1805,7 @@ int tty_connect(void)
                         /* Handle commands */
                         handle_command_sequence(input_char, &output_char, &forward);
 
-                        if ((option.hex_mode) && (forward))
+                        if ((option.input_mode == INPUT_MODE_HEX) && (forward))
                         {
                             if (!is_valid_hex(input_char))
                             {
@@ -1555,11 +1817,11 @@ int tty_connect(void)
 
                     if (forward)
                     {
-                        forward_to_tty(fd, output_char);
+                        forward_to_tty(device_fd, output_char);
                     }
                 }
 
-                tty_sync(fd);
+                tty_sync(device_fd);
             }
             else
             {
@@ -1568,16 +1830,29 @@ int tty_connect(void)
 
                 if (forward)
                 {
-                    forward_to_tty(fd, output_char);
+                    forward_to_tty(device_fd, output_char);
                 }
 
-                tty_sync(fd);
+                tty_sync(device_fd);
             }
         }
         else if (status == -1)
         {
+#if defined(__CYGWIN__)
+            // Happens when port unpluged
+            if (errno == EACCES)
+            {
+                goto error_read;
+            }
+#elif defined(__APPLE__)
+            if (errno == EBADF)
+            {
+                break; // tty_disconnect() will be naturally triggered by atexit()
+            }
+#else
             tio_error_printf("select() failed (%s)", strerror(errno));
             exit(EXIT_FAILURE);
+#endif
         }
         else
         {
