@@ -1,5 +1,5 @@
 /*
- * tio - a simple serial terminal I/O tool
+ * tio - a serial device I/O tool
  *
  * Copyright (c) 2014-2022  Martin Lund
  *
@@ -23,6 +23,9 @@
 
 #include <stdbool.h>
 
+#define LINE_HIGH true
+#define LINE_LOW false
+
 extern bool interactive_mode;
 extern bool map_i_nl_cr;
 extern bool map_i_cr_nl;
@@ -36,3 +39,7 @@ void tty_wait_for_device(void);
 void list_serial_devices(void);
 void tty_input_thread_create(void);
 void tty_input_thread_wait_ready(void);
+void tty_line_set(int fd, int mask, bool value);
+void tty_line_toggle(int fd, int mask);
+void tty_line_config(int mask, bool value);
+void tty_line_config_apply(void);
